@@ -20,17 +20,18 @@ from ipaddress import IPv4Address, IPv6Address
 
 import attr
 from aiorpcx import (
-    RPCSession, JSONRPCAutoDetect, JSONRPCConnection, serve_rs, serve_ws,
-    TaskGroup, handler_invocation, RPCError, Request, sleep, Event, ReplyAndDisconnect
+    RPCSession, JSONRPCAutoDetect, JSONRPCConnection, serve_rs, serve_ws, NewlineFramer,
+    TaskGroup, handler_invocation, RPCError, Request, sleep, Event, ReplyAndDisconnect,
+    timeout_after
 )
 import pylru
 
 import electrumx
 from electrumx.lib.merkle import MerkleCache
 from electrumx.lib.text import sessions_lines
-import electrumx.lib.util as util
-from electrumx.lib.hash import (sha256, hash_to_hex_str, hex_str_to_hash,
-                                HASHX_LEN, Base58Error)
+from electrumx.lib import util
+from electrumx.lib.hash import (sha256, hash_to_hex_str, hex_str_to_hash, HASHX_LEN, Base58Error,
+                                double_sha256)
 from electrumx.server.daemon import DaemonError
 from electrumx.server.peers import PeerManager
 
@@ -139,7 +140,6 @@ class SessionManager:
         self._merkle_hits = 0
         self.notified_height = None
         self.hsub_results = None
-        self._task_group = TaskGroup()
         self._sslc = None
         # Event triggered when electrumx is listening for incoming requests.
         self.server_listening = Event()
@@ -178,7 +178,7 @@ class SessionManager:
             host = None if service.host == 'all_interfaces' else str(service.host)
             try:
                 self.servers[service] = await serve(session_factory, host,
-                                                    service.port, ssl=sslc)
+                                                    service.port, ssl=sslc, reuse_address=True)
             except OSError as e:    # don't suppress CancelledError
                 self.logger.error(f'{kind} server failed to listen on {service.address}: {e}')
             else:
@@ -240,8 +240,9 @@ class SessionManager:
         if sessions:
             session_ids = ', '.join(str(session.session_id) for session in sessions)
             self.logger.info(f'{reason} session ids {session_ids}')
-            for session in sessions:
-                await self._task_group.spawn(session.close(force_after=force_after))
+            async with TaskGroup() as group:
+                for session in sessions:
+                    await group.spawn(session.close(force_after=force_after))
 
     async def _clear_stale_sessions(self):
         '''Cut off sessions that haven't done anything for 10 minutes.'''
@@ -296,7 +297,7 @@ class SessionManager:
             'coin': self.env.coin.__name__,
             'daemon': self.daemon.logged_url(),
             'daemon height': self.daemon.cached_height(),
-            'db height': self.db.db_height,
+            'db height': self.db.state.height,
             'db_flush_count': self.db.history.flush_count,
             'groups': len(self.session_groups),
             'history cache': cache_fmt.format(
@@ -365,7 +366,7 @@ class SessionManager:
         and record that as notified_height.
         '''
         # Paranoia: a reorg could race and leave db_height lower
-        height = min(height, self.db.db_height)
+        height = min(height, self.db.state.height)
         raw = await self.raw_header(height)
         self.hsub_results = {'hex': raw.hex(), 'height': height}
         self.notified_height = height
@@ -520,14 +521,6 @@ class SessionManager:
             except Base58Error:
                 pass
 
-            try:
-                script = coin.build_name_index_script(arg.encode("ascii"))
-                hashX = coin.name_hashX_from_script(script)
-                lines.append(f'Name: {arg}')
-                return hashX
-            except (AttributeError, UnicodeEncodeError):
-                pass
-
             return None
 
         for arg in items:
@@ -608,17 +601,21 @@ class SessionManager:
             for service in self.env.report_services:
                 self.logger.info(f'advertising service {service}')
             # Start notifications; initialize hsub_results
-            await notifications.start(self.db.db_height, self._notify_sessions)
+            await notifications.start(self.db.state.height, self._notify_sessions)
             await self._start_external_servers()
             # Peer discovery should start after the external servers
             # because we connect to ourself
-            async with self._task_group as group:
+            async with TaskGroup() as group:
                 await group.spawn(self.peer_mgr.discover_peers())
                 await group.spawn(self._clear_stale_sessions())
                 await group.spawn(self._handle_chain_reorgs())
                 await group.spawn(self._recalc_concurrency())
                 await group.spawn(self._log_sessions())
                 await group.spawn(self._manage_servers())
+
+                async for task in group:
+                    if not task.cancelled():
+                        task.result()
         finally:
             # Close servers then sessions
             self.logger.info('stopping servers')
@@ -638,7 +635,7 @@ class SessionManager:
             return 0
         return sum((group.cost() - session.cost) * group.weight for group in groups)
 
-    async def _merkle_branch(self, height, tx_hashes, tx_pos):
+    async def _merkle_branch(self, height, tx_hashes, tx_pos, tsc_format=False):
         tx_hash_count = len(tx_hashes)
         cost = tx_hash_count
 
@@ -655,12 +652,22 @@ class SessionManager:
                 merkle_cache = MerkleCache(self.db.merkle, tx_hashes_func)
                 self._merkle_cache[height] = merkle_cache
                 await merkle_cache.initialize(len(tx_hashes))
-            branch, _root = await merkle_cache.branch_and_root(tx_hash_count, tx_pos)
+            branch, root = await merkle_cache.branch_and_root(tx_hash_count, tx_pos,
+                                                              tsc_format=tsc_format)
         else:
-            branch, _root = self.db.merkle.branch_and_root(tx_hashes, tx_pos)
+            branch, root = self.db.merkle.branch_and_root(tx_hashes, tx_pos,
+                                                          tsc_format=tsc_format)
 
-        branch = [hash_to_hex_str(hash) for hash in branch]
-        return branch, cost / 2500
+        if tsc_format:
+            def converter(_hash):
+                if _hash == b"*":
+                    return _hash.decode()
+                else:
+                    return hash_to_hex_str(_hash)
+            branch = [converter(hash) for hash in branch]
+        else:
+            branch = [hash_to_hex_str(hash) for hash in branch]
+        return branch, root, cost / 2500
 
     async def merkle_branch_for_tx_hash(self, height, tx_hash):
         '''Return a triple (branch, tx_pos, cost).'''
@@ -671,9 +678,72 @@ class SessionManager:
             raise RPCError(
                 BAD_REQUEST, f'tx {hash_to_hex_str(tx_hash)} not in block at height {height:,d}'
             ) from None
-        branch, merkle_cost = await self._merkle_branch(height, tx_hashes, tx_pos)
+        branch, _root, merkle_cost = await self._merkle_branch(height, tx_hashes, tx_pos)
         return branch, tx_pos, tx_hashes_cost + merkle_cost
 
+    async def tsc_merkle_proof_for_tx_hash(self, height, tx_hash, txid_or_tx='txid',
+                                           target_type='block_hash'):
+        '''Return a pair (tsc_proof, cost) where tsc_proof is a dictionary with fields:
+            index - the position of the transaction
+            txOrId - either "txid" or "tx"
+            target - either "block_hash", "block_header" or "merkle_root"
+            nodes - the nodes in the merkle branch excluding the "target"'''
+
+        async def get_target(target_type):
+            try:
+                cost = 0.25
+                raw_header = await self.raw_header(height)
+                root_from_header = raw_header[36:36 + 32]
+                if target_type == "block_header":
+                    target = raw_header.hex()
+                elif target_type == "merkle_root":
+                    target = hash_to_hex_str(root_from_header)
+                else:  # target == block hash
+                    target = hash_to_hex_str(double_sha256(raw_header))
+            except ValueError:
+                raise RPCError(BAD_REQUEST, f'block header at height {height:,d} not found') \
+                    from None
+            return target, root_from_header, cost
+
+        def get_tx_position(tx_hash):
+            try:
+                tx_pos = tx_hashes.index(tx_hash)
+            except ValueError:
+                raise RPCError(BAD_REQUEST, f'tx {hash_to_hex_str(tx_hash)} not in block at height '
+                                            f'{height:,d}') from None
+            return tx_pos
+
+        async def get_txid_or_tx_field(tx_hash):
+            txid = hash_to_hex_str(tx_hash)
+            if txid_or_tx == "tx":
+                rawtx = await self.daemon_request('getrawtransaction', txid, False)
+                cost = 1.0
+                txid_or_tx_field = rawtx
+            else:
+                cost = 0.0
+                txid_or_tx_field = txid
+            return txid_or_tx_field, cost
+
+        tsc_proof = {}
+        tx_hashes, tx_hashes_cost = await self.tx_hashes_at_blockheight(height)
+        tx_pos = get_tx_position(tx_hash)
+        branch, root, merkle_cost = await self._merkle_branch(height, tx_hashes, tx_pos,
+                                                              tsc_format=True)
+
+        target, root_from_header, header_cost = await get_target(target_type)
+        # sanity check
+        if root != root_from_header:
+            raise RPCError(BAD_REQUEST, 'db error. Merkle root from cached block header does not '
+                                        'match the derived merkle root') from None
+
+        txid_or_tx_field, tx_fetch_cost = await get_txid_or_tx_field(tx_hash)
+
+        tsc_proof['index'] = tx_pos
+        tsc_proof['txid_or_tx'] = txid_or_tx_field
+        tsc_proof['target'] = target
+        tsc_proof['nodes'] = branch
+        return tsc_proof, tx_hashes_cost + merkle_cost + tx_fetch_cost + header_cost
+
     async def merkle_branch_for_tx_pos(self, height, tx_pos):
         '''Return a triple (branch, tx_hash_hex, cost).'''
         tx_hashes, tx_hashes_cost = await self.tx_hashes_at_blockheight(height)
@@ -683,7 +753,7 @@ class SessionManager:
             raise RPCError(
                 BAD_REQUEST, f'no tx at position {tx_pos:,d} in block at height {height:,d}'
             ) from None
-        branch, merkle_cost = await self._merkle_branch(height, tx_hashes, tx_pos)
+        branch, _root, merkle_cost = await self._merkle_branch(height, tx_hashes, tx_pos)
         return branch, hash_to_hex_str(tx_hash), tx_hashes_cost + merkle_cost
 
     async def tx_hashes_at_blockheight(self, height):
@@ -769,8 +839,9 @@ class SessionManager:
             for hashX in set(cache).intersection(touched):
                 del cache[hashX]
 
-        for session in self.sessions:
-            await self._task_group.spawn(session.notify, touched, height_changed)
+        async with TaskGroup() as group:
+            for session in self.sessions:
+                await group.spawn(session.notify, touched, height_changed)
 
     def _ip_addr_group_name(self, session):
         host = session.remote_address().host
@@ -850,13 +921,16 @@ class SessionBase(RPCSession):
         self.logger = util.ConnectionLogger(logger, context)
         self.logger.info(f'{self.kind} {self.remote_address_string()}, '
                          f'{self.session_mgr.session_count():,d} total')
-        self.recalc_concurrency()
         self.session_mgr.add_session(self)
+        self.recalc_concurrency()  # must be called after session_mgr.add_session
         self.request_handlers = {}
 
     async def notify(self, touched, height_changed):
         pass
 
+    def default_framer(self):
+        return NewlineFramer(max_size=self.env.max_recv)
+
     def remote_address_string(self, *, for_log=True):
         '''Returns the peer's IP address and port as a human-readable
         string, respecting anon logs if the output is for a log.'''
@@ -978,7 +1052,8 @@ class ElectrumX(SessionBase):
     async def notify(self, touched, height_changed):
         '''Wrap _notify_inner; websockets raises exceptions for unclear reasons.'''
         try:
-            await self._notify_inner(touched, height_changed)
+            async with timeout_after(30):
+                await self._notify_inner(touched, height_changed)
         except Exception:   # pylint:disable=W0703
             self.logger.exception('unexpected exception notifying client')
 
@@ -1160,7 +1235,7 @@ class ElectrumX(SessionBase):
         return self.unsubscribe_hashX(hashX) is not None
 
     async def _merkle_proof(self, cp_height, height):
-        max_height = self.db.db_height
+        max_height = self.db.state.height
         if not height <= cp_height <= max_height:
             raise RPCError(BAD_REQUEST,
                            f'require header height {height:,d} <= '
@@ -1365,6 +1440,32 @@ class ElectrumX(SessionBase):
 
         return {"block_height": height, "merkle": branch, "pos": tx_pos}
 
+    async def transaction_tsc_merkle(self, tx_hash, height, txid_or_tx='txid',
+                                     target_type='block_hash'):
+        '''Return the TSC merkle proof in JSON format to a confirmed transaction given its hash.
+        See: https://tsc.bitcoinassociation.net/standards/merkle-proof-standardised-format/.
+
+        tx_hash: the transaction hash as a hexadecimal string
+        include_tx: whether to include the full raw transaction in the response or txid.
+        target: options include: ('merkle_root', 'block_header', 'block_hash', 'None')
+        '''
+        tx_hash = assert_tx_hash(tx_hash)
+        height = non_negative_integer(height)
+
+        tsc_proof, cost = await self.session_mgr.tsc_merkle_proof_for_tx_hash(
+            height, tx_hash, txid_or_tx, target_type)
+        self.bump_cost(cost)
+
+        return {
+            "index": tsc_proof['index'],
+            "txOrId": tsc_proof['txid_or_tx'],
+            "target": tsc_proof['target'],
+            "nodes": tsc_proof['nodes'],  # "*" is used to represent duplicated hashes
+            "targetType": target_type,
+            "proofType": "branch",  # "tree" option is not supported by ElectrumX
+            "composite": False  # composite option is not supported by ElectrumX
+        }
+
     async def transaction_id_from_pos(self, height, tx_pos, merkle=False):
         '''Return the txid and optionally a merkle proof, given
         a block height and position in the block.
@@ -1411,6 +1512,7 @@ class ElectrumX(SessionBase):
             'blockchain.transaction.broadcast': self.transaction_broadcast,
             'blockchain.transaction.get': self.transaction_get,
             'blockchain.transaction.get_merkle': self.transaction_merkle,
+            'blockchain.transaction.get_tsc_merkle': self.transaction_tsc_merkle,
             'blockchain.transaction.id_from_pos': self.transaction_id_from_pos,
             'mempool.get_fee_histogram': self.compact_fee_histogram,
             'server.add_peer': self.add_peer,
